"""
wikinet
=======

wikinet is a Python package for the creation, manipulation, and study of networks of Wikipedia articles.

See https:// for complete documentation.
"""

from .dump import *
from .corpus import *
from .graph_container import *
from .persistent_homology import *
from .model import *
from .net import *
